from collections.abc import Sequence

from aioinject import Scoped, Singleton
from aioinject.validation._builtin import (
    ForbidDependency,
    all_dependencies_are_present,
)
from aioinject.validation._validate import validate_container
from aioinject.validation.abc import ContainerValidator


DEFAULT_VALIDATORS: Sequence[ContainerValidator] = [
    all_dependencies_are_present,
    ForbidDependency(
        dependant=lambda p: isinstance(p, Singleton),
        dependency=lambda p: isinstance(p, Scoped)
        and not isinstance(p, Singleton),
    ),
]

__all__ = [
    "DEFAULT_VALIDATORS",
    "ContainerValidator",
    "all_dependencies_are_present",
    "ForbidDependency",
    "validate_container",
]



import pytest

import aioinject


@pytest.fixture(scope="session")
def provided_value() -> int:
    return 42


class NumberService:
    async def get_number(self, number: int) -> int:
        return number


@pytest.fixture(scope="session")
def container(provided_value: int) -> aioinject.Container:
    container = aioinject.Container()
    container.register(aioinject.Object(provided_value))
    container.register(aioinject.Scoped(NumberService))
    return container



from __future__ import annotations

import contextvars
import inspect
from collections.abc import Callable, Coroutine, Iterable, Mapping
from contextvars import ContextVar
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    TypeAlias,
    TypeVar,
    Union,
    overload,
)

from typing_extensions import Self

from aioinject._store import InstanceStore, NotInCache
from aioinject.providers import Dependency, DependencyLifetime


if TYPE_CHECKING:
    from aioinject import Provider
    from aioinject.containers import Container

_T = TypeVar("_T")

_AnyCtx: TypeAlias = Union["InjectionContext", "SyncInjectionContext"]

context_var: ContextVar[_AnyCtx] = ContextVar("aioinject_context")
container_var: ContextVar[Container] = ContextVar("aioinject_container")


class _BaseInjectionContext:
    def __init__(
        self,
        container: Container,
        singletons: InstanceStore,
    ) -> None:
        self._container = container
        self._store = InstanceStore()
        self._singletons = singletons
        self._token: contextvars.Token[_AnyCtx] | None = None

    def _get_store(self, lifetime: DependencyLifetime) -> InstanceStore:
        if lifetime is DependencyLifetime.singleton:
            return self._singletons
        return self._store


class InjectionContext(_BaseInjectionContext):
    async def resolve(
        self,
        type_: type[_T],
    ) -> _T:
        provider = self._container.get_provider(type_)
        store = self._get_store(provider.lifetime)
        if (cached := store.get(provider)) is not NotInCache.sentinel:
            return cached

        dependencies = {}
        for dependency in provider.resolve_dependencies(
            self._container.type_context,
        ):
            dependencies[dependency.name] = await self.resolve(
                type_=dependency.type_,
            )

        if provider.lifetime is DependencyLifetime.singleton:
            async with store.lock(provider) as should_provide:
                if should_provide:
                    return await self._resolve(provider, store, dependencies)
                return store.get(  # type: ignore[return-value] # pragma: no cover
                    provider,
                )

        return await self._resolve(provider, store, dependencies)

    async def _resolve(
        self,
        provider: Provider[_T],
        store: InstanceStore,
        dependencies: Mapping[str, Any],
    ) -> _T:
        resolved = await provider.provide(dependencies)
        if provider.is_generator:
            resolved = await store.enter_context(resolved)
        store.add(provider, resolved)
        return resolved

    @overload
    async def execute(
        self,
        function: Callable[..., Coroutine[Any, Any, _T]],
        dependencies: Iterable[Dependency[object]],
        *args: Any,
        **kwargs: Any,
    ) -> _T:
        ...

    @overload
    async def execute(
        self,
        function: Callable[..., _T],
        dependencies: Iterable[Dependency[object]],
        *args: Any,
        **kwargs: Any,
    ) -> _T:
        ...

    async def execute(
        self,
        function: Callable[..., Coroutine[Any, Any, _T] | _T],
        dependencies: Iterable[Dependency[object]],
        *args: Any,
        **kwargs: Any,
    ) -> _T:
        resolved = {}
        for dependency in dependencies:
            if dependency.name in kwargs:
                continue

            resolved[dependency.name] = await self.resolve(
                type_=dependency.type_,
            )
        if inspect.iscoroutinefunction(function):
            return await function(*args, **kwargs, **resolved)
        return function(*args, **kwargs, **resolved)  # type: ignore[return-value]

    async def __aenter__(self) -> Self:
        self._token = context_var.set(self)
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        context_var.reset(self._token)  # type: ignore[arg-type]
        await self._store.__aexit__(exc_type, exc_val, exc_tb)


class SyncInjectionContext(_BaseInjectionContext):
    def resolve(
        self,
        type_: type[_T],
    ) -> _T:
        provider = self._container.get_provider(type_)
        store = self._get_store(provider.lifetime)
        if (cached := store.get(provider)) is not NotInCache.sentinel:
            return cached

        dependencies = {}
        for dependency in provider.resolve_dependencies():
            dependencies[dependency.name] = self.resolve(
                type_=dependency.type_,
            )

        if provider.lifetime is DependencyLifetime.singleton:
            with store.sync_lock(provider) as should_provide:
                if should_provide:
                    return self._resolve(provider, store, dependencies)
                return store.get(provider)  # type: ignore[return-value] # pragma: no cover

        return self._resolve(provider, store, dependencies)

    def _resolve(
        self,
        provider: Provider[_T],
        store: InstanceStore,
        dependencies: Mapping[str, Any],
    ) -> _T:
        resolved = provider.provide_sync(dependencies)
        if provider.is_generator:
            resolved = store.enter_sync_context(resolved)
        store.add(provider, resolved)
        return resolved

    def execute(
        self,
        function: Callable[..., _T],
        dependencies: Iterable[Dependency[object]],
        *args: Any,
        **kwargs: Any,
    ) -> _T:
        resolved = {}
        for dependency in dependencies:
            if dependency.name in kwargs:
                continue
            resolved[dependency.name] = self.resolve(type_=dependency.type_)
        return function(*args, **kwargs, **resolved)

    def __enter__(self) -> Self:
        self._token = context_var.set(self)
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        context_var.reset(self._token)  # type: ignore[arg-type]
        self._store.__exit__(exc_type, exc_val, exc_tb)



import contextlib
from collections.abc import AsyncIterator

import pytest
from litestar import Litestar

import aioinject
from aioinject import Container
from aioinject.ext.litestar import AioInjectPlugin


@pytest.mark.anyio
async def test_lifespan() -> None:
    number = 42

    shutdown = False

    @contextlib.asynccontextmanager
    async def dependency() -> AsyncIterator[int]:
        nonlocal shutdown
        yield number
        shutdown = True

    container = Container()
    container.register(aioinject.Singleton(dependency))

    app = Litestar(plugins=[AioInjectPlugin(container)])

    async with app.lifespan():
        async with container.context() as ctx:
            assert await ctx.resolve(int) == number

        assert shutdown is False
    assert shutdown is True



from typing import Annotated, NewType

import pytest

from aioinject import Container, Inject, Object, Scoped, inject, providers
from aioinject.context import container_var
from aioinject.decorators import InjectMethod


class _Session:
    pass


class _Service:
    def __init__(self, session: _Session) -> None:
        self.session = session


class _Interface:
    def __init__(self, session: _Session) -> None:
        self.session = session


@pytest.fixture
def container() -> Container:
    container = Container()
    container.register(providers.Scoped(_Session))
    container.register(providers.Scoped(_Service))
    return container


@inject
def _injectee(
    test: Annotated[_Session, Inject],
    test_no_cache: Annotated[_Session, Inject()],
) -> tuple[_Session, _Session]:
    return test, test_no_cache


def test_would_fail_without_active_context() -> None:
    with pytest.raises(LookupError):
        _injectee()  # type: ignore[call-arg]


def test_would_not_inject_without_inject_marker(container: Container) -> None:
    @inject
    def injectee(session: _Session) -> _Session:
        return session

    with container.sync_context(), pytest.raises(TypeError):
        injectee()  # type: ignore[call-arg]


def test_simple_inject(container: Container) -> None:
    with container.sync_context():
        session, *_ = _injectee()  # type: ignore[call-arg]
        assert isinstance(session, _Session)


@pytest.mark.anyio
async def test_simple_service(container: Container) -> None:
    with container.sync_context() as ctx:
        session = ctx.resolve(_Session)
        assert isinstance(session, _Session)

    async with container.context() as ctx:
        session = await ctx.resolve(_Session)
        assert isinstance(session, _Session)


@pytest.mark.anyio
async def test_retrieve_service_with_dependencies(
    container: Container,
) -> None:
    with container.sync_context() as ctx:
        service = ctx.resolve(_Service)
        assert isinstance(service, _Service)
        assert isinstance(service.session, _Session)

    async with container.context() as ctx:
        service = await ctx.resolve(_Service)
        assert isinstance(service, _Service)
        assert isinstance(service.session, _Session)


@pytest.mark.anyio
async def test_inject_using_container(
    container: Container,
) -> None:
    @inject(inject_method=InjectMethod.container)
    async def injectee(service: Annotated[_Service, Inject]) -> _Service:
        return service

    token = container_var.set(container)
    # This is fine
    coro = injectee()  # type: ignore[call-arg]
    assert isinstance(await coro, _Service)
    container_var.reset(token)


@pytest.mark.anyio
async def test_new_type() -> None:
    A = NewType("A", int)
    B = NewType("B", int)

    class Service:
        def __init__(self, a: A, b: B) -> None:
            self.a = a
            self.b = b

    container = Container()
    container.register(Object(1, type_=A))
    container.register(Object(2, type_=B))
    container.register(Scoped(Service))

    async with container.context() as ctx:
        service = await ctx.resolve(Service)
        assert service.a == 1
        assert service.b == 2  # noqa: PLR2004



import functools
from typing import Annotated

import pytest

import aioinject
from aioinject import Container, Inject
from aioinject.providers import collect_dependencies
from tests.context.conftest import _A, _C


def _dependant(
    a: Annotated[_A, Inject],
    c: Annotated[_C, Inject],
) -> tuple[_A, _C]:
    return a, c


async def _async_dependant(
    a: Annotated[_A, Inject],
    c: Annotated[_C, Inject],
) -> tuple[_A, _C]:
    return a, c


def test_execute_sync(container: Container) -> None:
    dependencies = list(collect_dependencies(_dependant))
    with container.sync_context() as ctx:
        a, c = ctx.execute(_dependant, dependencies)
        assert isinstance(a, _A)
        assert isinstance(c, _C)


def test_execute_sync_with_kwargs(container: Container) -> None:
    dependencies = list(collect_dependencies(_dependant))
    provided_a = _A()
    with container.sync_context() as ctx:
        a, c = ctx.execute(_dependant, dependencies, a=provided_a)
        assert a is provided_a
        assert isinstance(c, _C)


@pytest.mark.anyio
async def test_execute_async(container: Container) -> None:
    dependencies = list(collect_dependencies(_dependant))
    async with container.context() as ctx:
        a, c = await ctx.execute(_dependant, dependencies)
        assert isinstance(a, _A)
        assert isinstance(c, _C)


@pytest.mark.anyio
async def test_execute_async_with_kwargs(container: Container) -> None:
    dependencies = list(collect_dependencies(_dependant))
    provided_a = _A()
    async with container.context() as ctx:
        a, c = await ctx.execute(_dependant, dependencies, a=provided_a)
        assert isinstance(a, _A)
        assert isinstance(c, _C)


@pytest.mark.anyio
async def test_execute_async_coroutine(container: Container) -> None:
    dependencies = list(collect_dependencies(_async_dependant))
    async with container.context() as ctx:
        a, c = await ctx.execute(_async_dependant, dependencies)
        assert isinstance(a, _A)
        assert isinstance(c, _C)


@pytest.mark.anyio
async def test_provide_functools_partial() -> None:
    container = Container()
    container.register(
        aioinject.Scoped(functools.partial(str, 42), type_=str),
    )
    async with container.context() as ctx:
        assert await ctx.resolve(str) == "42"



import pytest

from aioinject import Container, Singleton


class _Test:
    pass


@pytest.fixture
def container() -> Container:
    container = Container()
    container.register(Singleton(_Test))
    return container


def test_identity(container: Container) -> None:
    with container.sync_context() as ctx:
        instance = ctx.resolve(_Test)

    with container.sync_context() as ctx:
        assert instance is ctx.resolve(_Test)


@pytest.mark.anyio
async def test_identity_async(container: Container) -> None:
    async with container.context() as ctx:
        instance = await ctx.resolve(_Test)

    async with container.context() as ctx:
        assert instance is await ctx.resolve(_Test)


@pytest.mark.anyio
async def test_should_not_execute_twice() -> None:
    count = 0

    async def func() -> int:
        nonlocal count
        count += 1
        return count

    container = Container()
    container.register(Singleton(func))
    for _ in range(5):
        async with container.context() as ctx:
            assert await ctx.resolve(int) == count
            assert count == 1



import enum
import functools
import inspect
from collections.abc import AsyncIterable, AsyncIterator, Callable, Coroutine
from typing import Any, ParamSpec, TypeVar, overload

from aioinject import InjectionContext, SyncInjectionContext
from aioinject.context import container_var, context_var
from aioinject.providers import collect_dependencies


_T = TypeVar("_T")
_P = ParamSpec("_P")
_ContextT = TypeVar("_ContextT", InjectionContext, SyncInjectionContext)


class InjectMethod(enum.Enum):
    container = enum.auto()
    context = enum.auto()


def _get_context(
    inject_method: InjectMethod,
    *,
    context_type: type[_ContextT],
) -> _ContextT:
    if inject_method is InjectMethod.container:
        container = container_var.get()
        if issubclass(context_type, InjectionContext):
            return container.context()
        return container.sync_context()
    return context_var.get()  # type: ignore[return-value]


def _wrap_async(
    function: Callable[_P, Coroutine[Any, Any, _T]],
    inject_method: InjectMethod,
) -> Callable[_P, Coroutine[Any, Any, _T]]:
    dependencies = list(collect_dependencies(function))

    @functools.wraps(function)
    async def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _T:
        context = _get_context(inject_method, context_type=InjectionContext)
        execute = context.execute(
            function,
            dependencies,
            *args,
            **kwargs,
        )

        if inject_method is InjectMethod.container:
            async with context:
                return await execute

        return await execute

    return wrapper


def _wrap_async_gen(
    function: Callable[_P, Coroutine[Any, Any, AsyncIterable[_T]]],
    inject_method: InjectMethod,
) -> Callable[_P, AsyncIterable[_T]]:
    wrapped = _wrap_async(function, inject_method)

    @functools.wraps(function)
    async def wrapper(
        *args: _P.args,
        **kwargs: _P.kwargs,
    ) -> AsyncIterator[_T]:
        async for element in await wrapped(*args, **kwargs):
            yield element

    return wrapper


def _wrap_sync(
    function: Callable[_P, _T],
    inject_method: InjectMethod,
) -> Callable[_P, _T]:
    dependencies = list(collect_dependencies(function))

    @functools.wraps(function)
    def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _T:
        context = _get_context(
            inject_method,
            context_type=SyncInjectionContext,
        )
        execute = functools.partial(
            context.execute,
            function,
            dependencies,
            *args,
            **kwargs,
        )

        if inject_method is InjectMethod.container:
            with context:
                return execute()

        return execute()

    return wrapper


@overload
def inject(
    func: Callable[_P, _T],
    *,
    inject_method: InjectMethod = InjectMethod.context,
) -> Callable[_P, _T]:
    ...


@overload
def inject(
    *,
    inject_method: InjectMethod = InjectMethod.context,
) -> Callable[[Callable[_P, _T]], Callable[_P, _T]]:
    ...


def inject(
    func: Callable[_P, _T] | None = None,
    *,
    inject_method: InjectMethod = InjectMethod.context,
) -> Callable[_P, _T] | Callable[[Callable[_P, _T]], Callable[_P, _T]]:
    def wrap(function: Callable[_P, _T]) -> Callable[_P, _T]:
        if inspect.iscoroutinefunction(function):
            return _wrap_async(
                function,
                inject_method=inject_method,
            )  # type: ignore[return-value]

        if inspect.isasyncgenfunction(function):
            return _wrap_async_gen(  # type: ignore[return-value]
                function,  # type: ignore[arg-type]
                inject_method=inject_method,
            )
        return _wrap_sync(function, inject_method=inject_method)

    if func is None:
        return wrap

    return wrap(func)



from collections.abc import Iterable

import aioinject
from aioinject.validation.abc import ContainerValidator
from aioinject.validation.error import (
    ContainerValidationErrorGroup,
)


def validate_container(
    container: aioinject.Container,
    validators: Iterable[ContainerValidator],
) -> None:
    for validator in validators:
        errors = validator(container)

        if errors:
            raise ContainerValidationErrorGroup(errors=errors)



import contextlib
from collections.abc import AsyncIterator, Iterator
from typing import Literal

import pytest

from aioinject import Object
from aioinject._store import InstanceStore, SingletonStore


pytestmark = [pytest.mark.anyio]

_NUMBER = 42


async def test_async_close() -> None:
    state: Literal["open", "closed"] | None = None

    @contextlib.asynccontextmanager
    async def ctx() -> AsyncIterator[int]:
        nonlocal state
        state = "open"
        yield _NUMBER
        state = "closed"

    store = InstanceStore()
    assert await store.enter_context(ctx()) == _NUMBER
    assert state == "open"

    store.close()
    assert state == "open"

    await store.aclose()
    assert state == "closed"  # type: ignore[comparison-overlap]


async def test_sync_close() -> None:
    state: Literal["open", "closed"] | None = None

    @contextlib.contextmanager
    def ctx() -> Iterator[int]:
        nonlocal state
        state = "open"
        yield _NUMBER
        state = "closed"

    store = InstanceStore()
    assert store.enter_sync_context(ctx()) == _NUMBER
    assert state == "open"

    await store.aclose()
    assert state == "open"

    store.close()
    assert state == "closed"  # type: ignore[comparison-overlap]


@pytest.mark.parametrize(
    "store_cls",
    [InstanceStore, SingletonStore],
)
async def test_lock(store_cls: type[InstanceStore]) -> None:
    store = store_cls()

    provider = Object(0)

    async with store.lock(provider) as should_provide:
        assert should_provide is True

    with store.sync_lock(provider) as should_provide:
        assert should_provide is True

    store.add(provider, 0)

    async with store.lock(provider) as should_provide:
        assert should_provide is False

    with store.sync_lock(provider) as should_provide:
        assert should_provide is False



from aioinject import Container, Object


class _A:
    pass


def test_provider_override() -> None:
    a = _A()
    container = Container()
    container.register(Object(a))
    with container.sync_context() as ctx:
        assert ctx.resolve(_A) is a

        a_overridden = _A()
        assert a_overridden is not a

    with container.sync_context() as ctx, container.override(
        Object(a_overridden),
    ):
        assert ctx.resolve(_A) is a_overridden is not a


def test_override_multiple_times() -> None:
    container = Container()
    with container.override(Object(1)):
        with container.override(Object(2)), container.sync_context() as ctx:
            assert ctx.resolve(int) == 2  # noqa: PLR2004

        with container.sync_context() as ctx:
            assert ctx.resolve(int) == 1



from collections.abc import Sequence
from typing import Any

import pytest

from aioinject import Container, Provider, Scoped, Singleton
from aioinject.validation import ForbidDependency, validate_container
from aioinject.validation.error import (
    ContainerValidationErrorGroup,
)


_VALIDATORS = [
    ForbidDependency(
        dependant=lambda p: isinstance(p, Singleton),
        dependency=lambda p: isinstance(p, Scoped)
        and not isinstance(p, Singleton),
    ),
]


def _str_dependency(number: int) -> str:
    return str(number)


@pytest.mark.parametrize(
    "providers",
    [
        [Scoped(int)],
        [Scoped(_str_dependency), Scoped(int)],
        [Scoped(_str_dependency), Singleton(int)],
        [Singleton(_str_dependency), Singleton(int)],
    ],
)
def test_ok(providers: Sequence[Provider[Any]]) -> None:
    container = Container()
    for provider in providers:
        container.register(provider)

    validate_container(container, _VALIDATORS)


@pytest.mark.parametrize(
    "providers",
    [
        [Singleton(_str_dependency), Scoped(int)],
    ],
)
def test_err(providers: Sequence[Provider[Any]]) -> None:
    container = Container()
    for provider in providers:
        container.register(provider)

    with pytest.raises(ContainerValidationErrorGroup):
        validate_container(container, _VALIDATORS)



import contextlib
from collections.abc import AsyncIterator, Generator, Iterator
from typing import Annotated, Any

import anyio
import pytest
from typing_extensions import Self

from aioinject import Object, Provider, Scoped, Singleton, providers
from aioinject.containers import Container
from aioinject.markers import Inject


pytestmark = [pytest.mark.anyio]


class _TestError(Exception):
    pass


class _Session:
    pass


class _Repository:
    def __init__(self, session: Annotated[_Session, Inject]) -> None:
        self.session = session


class _Service:
    def __init__(self, repository: Annotated[_Repository, Inject]) -> None:
        self.repository = repository


@pytest.fixture
def container() -> Container:
    container = Container()
    container.register(providers.Scoped(_Session))
    container.register(providers.Scoped(_Repository))
    container.register(providers.Scoped(_Service))
    return container


def test_can_instantiate_context(container: Container) -> None:
    assert container.context()


def test_can_retrieve_service(container: Container) -> None:
    with container.sync_context() as ctx:
        service = ctx.resolve(_Service)
        assert isinstance(service, _Service)
        assert isinstance(service.repository, _Repository)
        assert isinstance(service.repository.session, _Session)


def test_uses_cache(container: Container) -> None:
    with container.sync_context() as ctx:
        service = ctx.resolve(_Service)
        a, b, c = service, service.repository, service.repository.session

        service = ctx.resolve(_Service)
        assert a is service
        assert b is service.repository
        assert c is service.repository.session


def test_does_not_preserve_cache_if_recreated(container: Container) -> None:
    with container.sync_context() as ctx:
        service = ctx.resolve(_Service)

    with container.sync_context() as ctx:
        assert ctx.resolve(_Service) is not service


async def test_provide_async() -> None:
    class Test:
        pass

    container = Container()
    container.register(Scoped(Test))
    async with container.context() as ctx:
        instance = await ctx.resolve(Test)
        assert isinstance(instance, Test)


class _AwaitableCls:
    def __init__(self) -> None:
        self.awaited = False

    def __await__(self) -> Generator[Any, None, None]:
        self.awaited = True
        return anyio.sleep(0).__await__()


async def _async_awaitable() -> _AwaitableCls:
    return _AwaitableCls()


def _sync_awaitable() -> _AwaitableCls:
    return _AwaitableCls()


@pytest.mark.parametrize(
    "provider",
    [
        Scoped(_async_awaitable),  # type: ignore[arg-type]
        Scoped(_sync_awaitable),  # type: ignore[arg-type]
        Singleton(_async_awaitable),  # type: ignore[arg-type]
        Singleton(_sync_awaitable),  # type: ignore[arg-type]
    ],
)
async def test_should_not_execute_awaitable_classes(
    provider: Provider[_AwaitableCls],
) -> None:
    container = Container()
    container.register(provider)

    async with container.context() as ctx:
        resolved = await ctx.resolve(_AwaitableCls)
        assert isinstance(resolved, _AwaitableCls)
        assert not resolved.awaited


async def test_singleton_contextmanager_error() -> None:
    call_number = 0

    @contextlib.asynccontextmanager
    async def raises_error() -> AsyncIterator[int]:
        nonlocal call_number
        call_number += 1
        if call_number == 1:
            raise _TestError
        yield 42

    container = Container()
    container.register(Singleton(raises_error))

    with pytest.raises(_TestError):
        async with container.context() as ctx:
            await ctx.resolve(int)

    async with container.context() as ctx:
        await ctx.resolve(int)


async def test_returns_self() -> None:
    class Class:
        def __init__(self, number: str) -> None:
            self.number = number

        @classmethod
        async def self_classmethod(cls, number: int) -> Self:
            return cls(number=str(number))

        @classmethod
        @contextlib.asynccontextmanager
        async def async_context_classmethod(
            cls,
            number: int,
        ) -> AsyncIterator[Self]:
            yield cls(number=str(number))

        @classmethod
        @contextlib.contextmanager
        def sync_context_classmethod(cls, number: int) -> Iterator[Self]:
            yield cls(number=str(number))

    for factory in (
        Class.self_classmethod,
        Class.async_context_classmethod,
        Class.sync_context_classmethod,
    ):
        container = Container()
        container.register(Object(42))
        container.register(Scoped(factory))

        async with container.context() as ctx:
            instance = await ctx.resolve(Class)
            assert instance.number == "42"



import contextlib
from collections.abc import AsyncIterator, Generator, Iterator
from types import TracebackType
from unittest.mock import MagicMock

import pytest

from aioinject import Container, Scoped, providers
from tests.context.test_context import _Session


class _SyncContextManager:
    def __init__(self) -> None:
        self.mock = MagicMock()

    def __enter__(self) -> None:
        self.mock.open()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.mock.close()


class _AsyncContextManager:
    def __init__(self) -> None:
        self.mock = MagicMock()

    async def __aenter__(self) -> None:
        self.mock.open()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.mock.close()


def test_shutdowns_context_manager() -> None:
    mock = MagicMock()

    @contextlib.contextmanager
    def get_session() -> Iterator[_Session]:
        yield _Session()
        mock.close()

    container = Container()
    container.register(providers.Scoped(get_session))

    with container.sync_context() as ctx:
        session = ctx.resolve(_Session)
        assert isinstance(session, _Session)
        mock.close.assert_not_called()

    mock.close.assert_called()


def test_should_not_use_resolved_class_as_context_manager() -> None:
    container = Container()
    container.register(providers.Scoped(_SyncContextManager))

    with container.sync_context() as ctx:
        resolved = ctx.resolve(_SyncContextManager)
        resolved.mock.open.assert_not_called()
    resolved.mock.close.assert_not_called()


@pytest.mark.anyio
async def test_async_context_manager() -> None:
    mock = MagicMock()

    @contextlib.asynccontextmanager
    async def get_session() -> AsyncIterator[_Session]:
        mock.open()
        yield _Session()
        mock.close()

    container = Container()
    container.register(Scoped(get_session))
    async with container.context() as ctx:
        mock.open.assert_not_called()
        instance = await ctx.resolve(_Session)
        mock.open.assert_called_once()
        assert isinstance(instance, _Session)
    mock.close.assert_called_once()


@pytest.mark.anyio
async def test_async_context_would_use_sync_context_managers() -> None:
    mock = MagicMock()

    @contextlib.contextmanager
    def get_session() -> Generator[_Session, None, None]:
        mock.open()
        yield _Session()
        mock.close()

    container = Container()
    container.register(Scoped(get_session))
    async with container.context() as ctx:
        mock.open.assert_not_called()
        await ctx.resolve(_Session)
        mock.open.assert_called_once()
        mock.close.assert_not_called()
    mock.close.assert_called_once()


@pytest.mark.anyio
async def test_should_not_use_resolved_class_as_async_context_manager() -> None:
    container = Container()
    container.register(Scoped(_AsyncContextManager))
    async with container.context() as ctx:
        instance = await ctx.resolve(_AsyncContextManager)
        assert isinstance(instance, _AsyncContextManager)
    instance.mock.open.assert_not_called()
    instance.mock.close.assert_not_called()


def test_sync_context_manager_should_receive_exception() -> None:
    mock = MagicMock()

    class TestError(Exception):
        pass

    @contextlib.contextmanager
    def get_session() -> Iterator[_Session]:
        try:
            yield _Session()
        except TestError:
            mock.exception_happened()

    container = Container()
    container.register(Scoped(get_session))

    with (  # noqa: PT012
        pytest.raises(TestError),
        container.sync_context() as ctx,
    ):
        session = ctx.resolve(_Session)
        assert isinstance(session, _Session)
        mock.exception_happened.assert_not_called()
        raise TestError

    mock.exception_happened.asssert_called_once()



from __future__ import annotations

import contextlib
from collections.abc import AsyncIterator, Callable
from typing import TYPE_CHECKING, ParamSpec, TypeVar

from litestar import Litestar
from litestar.config.app import AppConfig
from litestar.middleware import MiddlewareProtocol
from litestar.plugins import InitPluginProtocol
from litestar.types import ASGIApp, Receive, Scope, Send

from aioinject import _utils, decorators


if TYPE_CHECKING:
    from aioinject.containers import Container

_T = TypeVar("_T")
_P = ParamSpec("_P")

_STATE_KEY = "__aioinject_container__"


def inject(function: Callable[_P, _T]) -> Callable[_P, _T]:
    wrapper = decorators.inject(
        function,
        inject_method=decorators.InjectMethod.context,
    )
    return _utils.clear_wrapper(wrapper)


class AioInjectMiddleware(MiddlewareProtocol):
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
    ) -> None:
        app: Litestar = scope["app"]
        container: Container = app.state[_STATE_KEY]
        async with container.context():
            await self.app(scope, receive, send)


class AioInjectPlugin(InitPluginProtocol):
    def __init__(self, container: Container) -> None:
        self.container = container

    @contextlib.asynccontextmanager
    async def _lifespan(
        self,
        _: Litestar,
    ) -> AsyncIterator[None]:
        async with self.container:
            yield

    def on_app_init(self, app_config: AppConfig) -> AppConfig:
        app_config.state[_STATE_KEY] = self.container
        app_config.middleware.append(AioInjectMiddleware)
        app_config.lifespan.append(self._lifespan)
        return app_config



import contextlib
from collections.abc import AsyncIterator

import pytest

from aioinject import Container
from aioinject.providers import Transient


class _Test:
    pass


@pytest.fixture
def container() -> Container:
    container = Container()
    container.register(Transient(_Test))
    return container


@pytest.mark.anyio
def test_identity(container: Container) -> None:
    with container.sync_context() as ctx:
        assert ctx.resolve(_Test) is not ctx.resolve(_Test)


@pytest.mark.anyio
async def test_identity_async(container: Container) -> None:
    async with container.context() as ctx:
        assert await ctx.resolve(_Test) is not await ctx.resolve(_Test)


@pytest.mark.anyio
async def test_should_close_transient_dependencies() -> None:
    count = 0

    @contextlib.asynccontextmanager
    async def dependency() -> AsyncIterator[int]:
        nonlocal count
        count += 1
        yield count
        count -= 1

    container = Container()
    container.register(Transient(dependency))

    max_count = 10
    async with container.context() as ctx:
        for i in range(1, max_count + 1):
            assert await ctx.resolve(int) == count == i
        assert count == max_count
    assert count == 0
